use std::{collections::HashMap};

use crate::parser::ASTPart;

#[derive(Debug, Clone)]
pub struct Operation {
    pub opcode: u8,
    pub arg1: Option<u8>,
    pub arg2: Option<i64>,
    pub arg3: Option<u8>,
}

#[derive(Debug, Clone)]
struct RegisterState {
    id: u8,
    used: bool,
    variable: u32,
    last_used: usize,
}

#[derive(Debug, Clone)]
struct AllocateResult {
    register: u8,
    unbind_before: bool,
}

fn get_register_by_id(registers: &Vec<RegisterState>, id: u8) -> RegisterState {
    for register in registers {
        if register.id == id {
            return register.clone();
        }
    }
    panic!("Register with id {} not found", id);
}

fn get_register_by_variable(registers: &Vec<RegisterState>, variable: u32) -> RegisterState {
    for register in registers {
        if register.variable == variable {
            return register.clone();
        }
    }
    panic!("Register with variable {} not found", variable);
}

fn set_register(registers: &mut Vec<RegisterState>, new_state: RegisterState) {
    if new_state.id == 0 {
        return;
    }
    for i in 0..registers.len() {
        if registers[i].id == new_state.id {
            registers[i] = new_state;
            return;
        }
    }
}

fn allocate_register(registers: &Vec<RegisterState>) -> AllocateResult {
    for register in registers {
        if register.id != 0 && !register.used {
            return AllocateResult {
                register: register.id,
                unbind_before: false,
            };
        }
    }
    //Find the one that wasn't used for the longest time
    let mut oldest: usize = 0;
    let mut oldest_register: u8 = 0;
    for register in registers {
        if ((register.last_used < oldest) || (oldest == 0)) && register.id != 0 {
            oldest = register.last_used;
            oldest_register = register.id;
        }
    }
    return AllocateResult {
        register: oldest_register,
        unbind_before: true,
    };
}

fn garbage_collect_registers(registers: &mut Vec<RegisterState>) {
    for register in registers {
        if register.used && register.variable == 0 {
            register.used = false;
            register.last_used = 0;
        }
    }
}

fn do_ast_op(ast_op: ASTPart, op_count: &mut usize, ops: &mut Vec<Operation>, var_ids: &mut HashMap<String, u32>, next_var_id: &mut u32, strings: &mut HashMap<u32, String>, next_string_id: &mut u32, registers: &mut Vec<RegisterState>) -> u8 {
    *op_count += 1;
    match ast_op {
        ASTPart::Number(num) => {
            let reg = allocate_register(registers);
            if reg.unbind_before {
                ops.push(Operation { opcode: 8, arg1: Some(reg.register), arg2: None, arg3: None });
            }
            ops.push(Operation { opcode: 3, arg1: Some(reg.register), arg2: Some(num.value), arg3: None });
            set_register(registers, RegisterState { id: reg.register, used: true, variable: 0, last_used: *op_count });
            return reg.register;
        },
        ASTPart::String(str) => {
            let str_id = *next_string_id;
            strings.insert(str_id, str.value);
            *next_string_id += 1;
            let reg = allocate_register(registers);
            if reg.unbind_before {
                ops.push(Operation { opcode: 8, arg1: Some(reg.register), arg2: None, arg3: None });
            }
            ops.push(Operation { opcode: 1, arg1: Some(reg.register), arg2: Some(str_id as i64), arg3: None });
            set_register(registers, RegisterState { id: reg.register, used: true, variable: 0, last_used: *op_count });
            return reg.register;
        },
        ASTPart::Operation(op) => {
            let left_reg = do_ast_op(*op.left, op_count, ops, var_ids, next_var_id, strings, next_string_id, registers);
            let right_reg = do_ast_op(*op.right, op_count, ops, var_ids, next_var_id, strings, next_string_id, registers);
            let reg = allocate_register(registers);
            if reg.unbind_before {
                ops.push(Operation { opcode: 8, arg1: Some(reg.register), arg2: None, arg3: None });
                set_register(registers, RegisterState { id: reg.register, used: true, variable: 0, last_used: *op_count });
            }
            let opcode = match op.operator.as_str() {
                "+" => 10,
                "-" => 11,
                "*" => 12,
                "/" => 13,
                "^" => 14,
                "%" => 15,
                "&" => 16,
                "|" => 17,
                "==" => 18,
                "!=" => 19,
                ">" => 20,
                ">=" => 21,
                "<" => 22,
                "<=" => 23,
                _ => panic!("Unknown operator {}", op.operator),
            };
            ops.push(Operation { opcode, arg1: Some(left_reg), arg2: Some(right_reg as i64), arg3: Some(reg.register) });
            set_register(registers, RegisterState { id: reg.register, used: true, variable: 0, last_used: *op_count });
            return reg.register;
        },
        ASTPart::Assigment(asign) => {
            if var_ids.contains_key(&asign.variable) {
                panic!("Variable {} already exists", asign.variable);
            }
            var_ids.insert(asign.variable.clone(), *next_var_id);
            *next_var_id += 1;
            let reg = do_ast_op(*asign.value, op_count, ops, var_ids, next_var_id, strings, next_string_id, registers);
            ops.push(Operation { opcode: 7, arg1: Some(reg), arg2: Some(var_ids[&asign.variable.clone()] as i64), arg3: None });
            set_register(registers, RegisterState { id: reg, used: true, variable: var_ids[&asign.variable.clone()], last_used: *op_count });
            garbage_collect_registers(registers);
        },
        _ => {}
    }
    return 0;
}

pub fn compile(ast: Vec<ASTPart>) -> Vec<u8> {
    let mut ops: Vec<Operation> = vec![];
    
    let mut var_ids: HashMap<String, u32> = HashMap::new();
    let mut next_var_id: u32 = 1;
    let mut strings: HashMap<u32, String> = HashMap::new();
    let mut next_string_id: u32 = 1;
    let mut registers: Vec<RegisterState> = vec![];
    for i in 0..17 {
        registers.push(RegisterState {
            id: i as u8,
            used: false,
            variable: 0,
            last_used: 0,
        });
    }

    let mut op_count = 0;
    for ast_op in ast {
        do_ast_op(ast_op, &mut op_count, &mut ops, &mut var_ids, &mut next_var_id, &mut strings, &mut next_string_id, &mut registers);
        println!("Operations: {:?}\n", ops);
        println!("Registers: {:?}\n", registers);
        println!("Variable IDs: {:?}", var_ids);
        println!("Strings: {:?}", strings);
        println!("==========================");
    }
    return vec![];
}